﻿local GCD = LibStub("AceAddon-3.0"):NewAddon("GCD", "AceConsole-3.0", "AceEvent-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("GCD")

local GetTime = GetTime

local GCDBar
local starttime, duration, spellid, warned, caststarttime, castendtime, castduration, castinterrupted, castlatency, castsent, iscasting, isgcd

local Frame1, Frame2, Frame3, Frame4
local spell1, spell2, spell3, spell4

local sin, cos, rad, ceil, floor = math.sin, math.cos, math.rad, math.ceil, math.floor

local class

local BOOKTYPE_SPELL = BOOKTYPE_SPELL

GCD.options = {
	desc = "GCD",
	type = "group",
	args = {
		spell = {
			name = L["Spell"],
			desc = L["Set Spell to track GCD"],
			type = "input",
			set = function(info,val)
				GCD.db.profile.spell = strlower(val)
				GCD:SPELLS_CHANGED() end,
		},
		casttime = {
			name = L["Casttime"],
			desc = L["Toggle indicator for Casttime + Latency"],
			type = "toggle",
			get = function(info) return GCD.db.profile.casttime end,
			set = function(info,val)
				GCD.db.profile.casttime = val
				if val then
					GCD:RegisterEvent("UNIT_SPELLCAST_SENT")
					GCD:RegisterEvent("UNIT_SPELLCAST_START")
					GCD:RegisterEvent("UNIT_SPELLCAST_STOP")
				end
				GCD:Print(L["Casttime"].." => "..tostring(val)) end,
		},
		spellcds = {
			name = L["SpellCooldowns"],
			desc = L["Toggle cooldown icons for spells"],
			type = "toggle",
			get = function(info) return GCD.db.profile.spellcds end,
			set = function(info,val)
				GCD.db.profile.spellcds = val
				if val then
					GCD:ACTIONBAR_UPDATE_COOLDOWN()
				else
					Frame1:Hide()
					Frame2:Hide()
					Frame3:Hide()
					Frame4:Hide()
				end
				GCD:Print(L["SpellCooldowns"].." => "..tostring(val)) end,
		},
	},
}

local function OnUpdate(self, elapsed)
	if not iscasting and not isgcd then
		GCDBar:Hide()
	end
	local x, y = GetCursorPosition(UIParent);
	self:SetPoint('CENTER', UIParent, 'BOTTOMLEFT', (x+self.xOff)/self:GetEffectiveScale(), (y+self.yOff)/self:GetEffectiveScale())
		
	if self.spark then
		self.texture.timer = self.texture.timer + elapsed;
		if ( self.texture.timer > 0.02 ) then
			self.texture.hAngle = self.texture.hAngle + 0.5;
			self.texture.s = sin(rad(self.texture.hAngle));
			self.texture.c = cos(rad(self.texture.hAngle));
			self.texture:SetTexCoord(	0.5-self.texture.s, 0.5+self.texture.c,
						0.5+self.texture.c, 0.5+self.texture.s,
						0.5-self.texture.c, 0.5-self.texture.s,
						0.5+self.texture.s, 0.5-self.texture.c);
			self.texture.timer = 0;
		end
		if isgcd then
			local gcdperc = (GetTime() - self.starttime) / self.duration
			gcdperc = gcdperc < 1 and gcdperc or 1			
			local angle = rad(90-gcdperc*360)
			local x, y = cos(angle), sin(angle)
			self.spark:SetPoint('CENTER', self, 'CENTER',x*25,y*25)
			self.spark:Show()
			if gcdperc >= 1 then
				isgcd = false
				self.spark:Hide()
			end
		else
			self.spark:Hide()
		end
		if iscasting and GCD.db.profile.casttime then
			local castperc = (1000 * GetTime() - caststarttime) / castduration
			castperc = castperc < 1 and castperc or 1
			if castperc > 1-castlatency then
				self.castspark:SetVertexColor(1,0,0)
			elseif castinterrupted then
				self.castspark:SetVertexColor(1,1,0)
			else
				self.castspark:SetVertexColor(0,1,0)
			end
			local angle = rad(90-castperc*360)
			local x, y = cos(angle), sin(angle)
			self.castspark:SetPoint('CENTER', self, 'CENTER',x*30,y*30)
			self.castspark:Show()
			if castperc >= 1 then
				iscasting = false
				self.castspark:Hide()
			end
		else
			self.castspark:Hide()
		end
	elseif self.spell and GCD.db.profile.spellcds then
		local perc = (GetTime() - self.starttime) / self.duration
		self.texture:SetVertexColor(1.0,perc,perc)
		local dur = floor((self.duration - (GetTime() - self.starttime))*10)/10
		if dur > 99 then
			dur = ''
		elseif dur > 10 then
			dur = ceil(dur)
		end
		self.cdtext:SetText(dur)
		if perc >= 1 then
			self:Hide()
		end
	end
end

local function OnHide(self)
	self:SetScript('OnUpdate', nil)
end

local function OnShow(self)
	self:SetScript('OnUpdate', OnUpdate)
end

local defaults = {
	char = {
		spell = '',
		spellcds = false,
		casttime = true,
	}
}

local function GetSpellID(spellname)
	local spellid = nil
	for tab = 1, 4 do
		local _, _, offset, numSpells = GetSpellTabInfo(tab)
		for i = (1+offset), (offset+numSpells) do
			local spell = GetSpellName(i, BOOKTYPE_SPELL)
			if strlower(spell) == strlower(spellname) then
				spellid = i
				break
			end
		end
	end
	return spellid;
end

function GCD:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("GCDDB", defaults, "char")
	LibStub("AceConfig-3.0"):RegisterOptionsTable("GCD", self.options)
	LibStub("AceConfigCmd-3.0"):CreateChatCommand("gcd", "GCD")
end

function GCD:OnEnable()
	_, class = UnitClass("player")
	if GCD.db.profile.casttime then
		self:RegisterEvent("UNIT_SPELLCAST_SENT")
		self:RegisterEvent("UNIT_SPELLCAST_START")
		self:RegisterEvent("UNIT_SPELLCAST_STOP")
	end
	self:RegisterEvent("ACTIONBAR_UPDATE_COOLDOWN")
	self:RegisterEvent("SPELLS_CHANGED")
	if not GCDBar then
		GCDBar = CreateFrame('Frame', 'GCDBar', UIParent)
		GCDBar:ClearAllPoints()
		GCDBar:SetFrameStrata('HIGH')
		GCDBar:SetScript('OnShow', OnShow)
		GCDBar:SetScript('OnHide', OnHide)
		GCDBar:SetWidth(64)
		GCDBar:SetHeight(64)
		GCDBar:Hide()
		GCDBar.xOff = 10
		GCDBar.yOff = -10
		
		GCDBar.texture = GCDBar:CreateTexture(nil, 'ARTWORK')
		GCDBar.texture:SetTexture("Spells\\AURARUNE8")
		GCDBar.texture:SetVertexColor(0,1,0,0.5)
		GCDBar.texture:SetBlendMode('ADD')
		GCDBar.texture:SetWidth(80)
		GCDBar.texture:SetHeight(80)
		GCDBar.texture.timer = 0;
		GCDBar.texture.hAngle = 0;
		GCDBar.texture:SetPoint('CENTER', GCDBar, 'CENTER')
		
		GCDBar.spark = GCDBar:CreateTexture(nil, 'DIALOG')
		GCDBar.spark:SetTexture("Spells\\AURA_01")
		GCDBar.spark:SetBlendMode('ADD')
		GCDBar.spark:SetWidth(30)
		GCDBar.spark:SetHeight(30)
		
		GCDBar.castspark = GCDBar:CreateTexture(nil, 'DIALOG')
		GCDBar.castspark:SetTexture("Spells\\AURA_01")
		GCDBar.castspark:SetVertexColor(0,1,0)
		GCDBar.castspark:SetBlendMode('ADD')
		GCDBar.castspark:SetWidth(30)
		GCDBar.castspark:SetHeight(30)
	end
	if class == "DRUID" or class == "PALADIN" then
		if not Frame1 then
			Frame1 = CreateFrame('Frame', 'GCDFrame1', UIParent)
			Frame1:ClearAllPoints()
			Frame1:SetFrameStrata('HIGH')
			Frame1:SetWidth(1)
			Frame1:SetHeight(1)
			Frame1:SetScript('OnShow', OnShow)
			Frame1:SetScript('OnHide', OnHide)
			Frame1:Hide()
			Frame1.xOff = -20
			Frame1.yOff = 20
			Frame1.spell = true
			
			Frame1.texture = Frame1:CreateTexture(nil, 'ARTWORK')
			if class == "DRUID" then
				Frame1.texture:SetTexture("Interface\\Icons\\Ability_Druid_Flourish") -- Wild Growth
			elseif class == "PALADIN" then
				Frame1.texture:SetTexture("Interface\\Icons\\Spell_Holy_SearingLight") -- Holy Shock
			end
			Frame1.texture:SetVertexColor(1,0,0)
			Frame1.texture:SetWidth(15*Frame1:GetEffectiveScale())
			Frame1.texture:SetHeight(15*Frame1:GetEffectiveScale())
			Frame1.texture:SetPoint('CENTER', Frame1, 'CENTER')
			
			Frame1.cdtext = Frame1:CreateFontString()
			Frame1.cdtext:SetFont("Fonts\\FRIZQT__.TTF", 11, "OUTLINE, MONOCHROME")
			Frame1.cdtext:SetPoint('CENTER', Frame1, 'CENTER', 0, 1)
		end
		if not Frame2 then
			Frame2 = CreateFrame('Frame', 'GCDFrame2', UIParent)
			Frame2:ClearAllPoints()
			Frame2:SetFrameStrata('HIGH')
			Frame2:SetWidth(1)
			Frame2:SetHeight(1)
			Frame2:SetScript('OnShow', OnShow)
			Frame2:SetScript('OnHide', OnHide)
			Frame2:Hide()
			Frame2.xOff = 40
			Frame2.yOff = 20
			Frame2.spell = true
			
			Frame2.texture = Frame2:CreateTexture(nil, 'ARTWORK')
			if class == "DRUID" then
				Frame2.texture:SetTexture("Interface\\Icons\\INV_Relics_IdolOfRejuvenation") -- Rejuvenation
			end
			Frame2.texture:SetVertexColor(1,0,0)
			Frame2.texture:SetWidth(15*Frame1:GetEffectiveScale())
			Frame2.texture:SetHeight(15*Frame1:GetEffectiveScale())
			Frame2.texture:SetPoint('CENTER', Frame2, 'CENTER')
			
			Frame2.cdtext = Frame2:CreateFontString()
			Frame2.cdtext:SetFont("Fonts\\FRIZQT__.TTF", 11, "OUTLINE, MONOCHROME")
			Frame2.cdtext:SetPoint('CENTER', Frame2, 'CENTER', 0, 1)
		end
		if not Frame3 then
			Frame3 = CreateFrame('Frame', 'GCDFrame3', UIParent)
			Frame3:ClearAllPoints()
			Frame3:SetFrameStrata('HIGH')
			Frame3:SetWidth(1)
			Frame3:SetHeight(1)
			Frame3:SetScript('OnShow', OnShow)
			Frame3:SetScript('OnHide', OnHide)
			Frame3:Hide()
			Frame3.xOff = -20
			Frame3.yOff = -40
			Frame3.spell = true
			
			Frame3.texture = Frame3:CreateTexture(nil, 'ARTWORK')
			if class == "DRUID" then
				Frame3.texture:SetTexture("Interface\\Icons\\Spell_Nature_RavenForm")
			end
			Frame3.texture:SetVertexColor(1,0,0)
			Frame3.texture:SetWidth(15*Frame1:GetEffectiveScale())
			Frame3.texture:SetHeight(15*Frame1:GetEffectiveScale())
			Frame3.texture:SetPoint('CENTER', Frame3, 'CENTER')
			
			Frame3.cdtext = Frame3:CreateFontString()
			Frame3.cdtext:SetFont("Fonts\\FRIZQT__.TTF", 11, "OUTLINE, MONOCHROME")
			Frame3.cdtext:SetPoint('CENTER', Frame3, 'CENTER', 0, 1)
		end
		if not Frame4 then
			Frame4 = CreateFrame('Frame', 'GCDFrame4', UIParent)
			Frame4:ClearAllPoints()
			Frame4:SetFrameStrata('HIGH')
			Frame4:SetWidth(1)
			Frame4:SetHeight(1)
			Frame4:SetScript('OnShow', OnShow)
			Frame4:SetScript('OnHide', OnHide)
			Frame4:Hide()
			Frame4.xOff = 40
			Frame4.yOff = -40
			Frame4.spell = true
			
			Frame4.texture = Frame4:CreateTexture(nil, 'ARTWORK')
			--Frame4.texture:SetTexture("Interface\\Icons\\Spell_Nature_RavenForm")
			Frame4.texture:SetVertexColor(1,0,0)
			Frame4.texture:SetWidth(15*Frame1:GetEffectiveScale())
			Frame4.texture:SetHeight(15*Frame1:GetEffectiveScale())
			Frame4.texture:SetPoint('CENTER', Frame4, 'CENTER')
			
			Frame4.cdtext = Frame4:CreateFontString()
			Frame4.cdtext:SetFont("Fonts\\FRIZQT__.TTF", 11, "OUTLINE, MONOCHROME")
			Frame4.cdtext:SetPoint('CENTER', Frame4, 'CENTER', 0, 1)
		end
	end
	self:SPELLS_CHANGED()
end

function GCD:OnDisable()
	GCDBar:Hide()
	self:UnregisterAllEvents()
	GCDBar = nil
	Frame1 = nil
	Frame2 = nil
	Frame3 = nil
	Frame4 = nil
end

function GCD:ACTIONBAR_UPDATE_COOLDOWN()
	if spellid then
		local start, dur = GetSpellCooldown(spellid, BOOKTYPE_SPELL)
		if dur > 0 and dur <= 2 then
			isgcd = true
			GCDBar.starttime = start
			GCDBar.duration = dur
			GCDBar:Show()
		end
		if self.db.profile.spellcds then
			local _, odur = GetSpellCooldown(spellid, BOOKTYPE_SPELL)
			if spell1 then
				local start, dur = GetSpellCooldown(spell1, BOOKTYPE_SPELL)
				if dur > odur then
					Frame1.starttime = start
					Frame1.duration = dur
					Frame1:Show()
				end
			end
			if spell2 then
				local start, dur = GetSpellCooldown(spell2, BOOKTYPE_SPELL)
				if dur > odur then
					Frame2.starttime = start
					Frame2.duration = dur
					Frame2:Show()
				end
			end
			if spell3 then
				local start, dur = GetSpellCooldown(spell3, BOOKTYPE_SPELL)
				if dur > odur then
					Frame3.starttime = start
					Frame3.duration = dur
					Frame3:Show()
				end
			end
			if spell4 then
				local start, dur = GetSpellCooldown(spell4, BOOKTYPE_SPELL)
				if dur > odur then
					Frame4.starttime = start
					Frame4.duration = dur
					Frame4:Show()
				end
			end
		end
	end
end

function GCD:UNIT_SPELLCAST_SENT(unit)
	castsent = GetTime()*1000
end

function GCD:UNIT_SPELLCAST_START(object, unit)
	if unit ~= 'player' then
		return
	end
	iscasting = true
	castinterrupted = false
	_, _, _, _, caststarttime, castendtime = UnitCastingInfo(unit)
	local sendlag= (castsent and castsent > 0) and GetTime()*1000-castsent or 0
	castduration=castendtime-caststarttime
	sendlag = sendlag > castduration and castduration or sendlag
	castlatency = sendlag/castduration
	GCDBar:Show()
end

function GCD:UNIT_SPELLCAST_STOP(unit)
	if unit ~= 'player' then
		return
	end
  iscasting = false
end

function GCD:SPELLS_CHANGED()
	if GCDBar then
		if not self.db.profile.spell then
			if not warned then
				self:Print(L["No spell set! Set one with \"/gcd spell <spellname>\""])
				warned = true
			end
		else
			spellid = GetSpellID(self.db.profile.spell)
			if not spellid then
				self:Print(L["Invalid spell set! Set new with \"/gcd spell <spellname>\""])
				self.db.profile.spell = nil
			end
		end
		if class == "DRUID" then
			spell1 = GetSpellID(L["Wild Growth"])
			spell2 = GetSpellID(L["Swiftmend"])
			spell3 = GetSpellID(L["Nature's Swiftness"])
		elseif class == "PALADIN" then
			spell1 = GetSpellID(L["Holy Shock"])
		end
	end
end
